package net.sf.xanswer.storage.map;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashSet;

public abstract class PersistentMapImplBase implements PersistentMap {
    
    protected abstract String getInstanceName();
    
    protected abstract void createDatabase(File location, int pageSize, int buffersCount) throws PersistentException;
    protected abstract void openDatabase() throws PersistentException;
    protected abstract void closeDatabase() throws PersistentException;
    
    protected abstract void putBytes(String key, byte[] values) throws PersistentException;
    protected abstract byte[] getBytes(String key) throws PersistentException;
    protected abstract void removeBytes(String key) throws PersistentException;

    protected abstract StringIterator getIterator() throws PersistentException;
    protected abstract long getNumberOfRecords() throws PersistentException;
    
    private final PersistentFactory myFactory;
    private File myLocationFile;
    private boolean isOpened = false;
    private long myLastModified;

    protected PersistentMapImplBase(PersistentFactory factory) {
        myFactory = factory;
    }

    public PersistentFactory getFactory() {
    	return myFactory;
    }

    public String toString() {
        return getInstanceName()+" database, location "+getLocation();
    }

    public PersistentEntry getEntry(String key){
    	return new PersistentEntryImpl(key); // nested class
    }

    // TODO: implement iterator accessing BTree node-by-node
    // current problem: only one iterator may exist at one time moment
    
    // TODO: ensure save removing from the tree during iteration
    // current problem: iterator behavior is unpredictable if
    // a deletion occurs! For now we are firmly tied with the tree
    // structure and consistence while iterating it.
    public PersistentIterator iterator() throws PersistentException {
        return new PersistentIteratorImpl(getIterator()); // nested class
    }

    public void create(String location) throws PersistentException {
		create(location, DEFAULT_PAGE_SIZE, DEFAULT_BUFFERS_COUNT);
   	}

    public void create(String location, int pageSize, int buffersCount) throws PersistentException {
   		if (location == null) throw new NullPointerException("Location of the db is null!");
		myLocationFile = new File(location);
		if ((pageSize <= 0) || (buffersCount <= 0)) {
		    throw new IllegalArgumentException("I got wrong settings (<=0): pageSize="+pageSize+
		            ", buffersCount="+buffersCount);
		}
		createDatabase(myLocationFile, pageSize, buffersCount);		    
		if (myLocationFile.isDirectory()) {
		    myLastModified = System.currentTimeMillis();
		} else {
		    myLastModified = myLocationFile.lastModified();		    
		}
   	}

    public void open() throws PersistentException {
    	if (!isCreated()) throw new IllegalStateException("Trying to open not existing "+toString());
    	openDatabase();
       	isOpened = true;
    }


    public void close() throws PersistentException {
    	if (!isCreated()) throw new IllegalStateException("Trying to close not existing "+toString());
    	closeDatabase();
       	isOpened = false;
    }

    public String getLocation() {
        String result = "location is null";
        if (myLocationFile != null) {
            result = myLocationFile.getAbsolutePath();
        }
        return result;
    }

    public long getLastModificationTime() {
    	if (!isCreated()) throw new RuntimeException("I'm not created!!!");
    	return myLastModified;
    }
    
    public long size() throws PersistentException {
        checkState();
        return getNumberOfRecords();
    }
    
    public String getName() {
        String result = myLocationFile.getName();
        if (!myLocationFile.isDirectory()) {
            // remove ".db" file extension - and we got a name for referensing this db with dbFactory
            result = result.substring(0, result.lastIndexOf("."));
        }
        return result;
    }
    
    private void putCollection(String key, Collection/*<String>*/ value) throws PersistentException {
        checkState();
        if (key==null || value==null) {
            throw new IllegalArgumentException("key and value are expected to be not null");
        }
        byte[] bytes = getSerializedCollection(value);
        if ((bytes != null) && (bytes.length != 0)) {
            putBytes(key, bytes);            
            modified();
        }
    }
    
    private Collection/*<String>*/ getCollection(String key) throws PersistentException {
        if (key==null) {
            throw new IllegalArgumentException("Key is expected to be not null");
        }
        checkState();
        Collection/*<String>*/ result = null;
        byte[] bytes = getBytes(key);
        // in my subclasses  new byte[0] is returned if nothing found
        if ((bytes != null) && (bytes.length != 0)) {
            result = getDeserializedCollection(bytes);           
        }
        return result;
    }
    
    private void removeCollection(String key) throws PersistentException {
        if (key==null) {
            throw new IllegalArgumentException("Key is expected to be not null");
        }
        checkState();
        if (getCollection(key) != null) {
            removeBytes(key);
            modified();
        }
    }
    
    // Some private auxilary helpers
    
    private byte[] getSerializedCollection(Collection/*<?>*/ value) throws PersistentException {
        byte[] result = new byte[0];
   	    try {
   	        ByteArrayOutputStream bos = new ByteArrayOutputStream();
   	        ObjectOutputStream out = new ObjectOutputStream(bos);
	        out.writeObject(value);
   	        result = bos.toByteArray();
   	        out.close();
   	        bos.close(); 	        
   	    } catch (IOException e) {
   	        throw new PersistentException(e);
   	    }
    	return result;
    }
	
    private Collection/*<?>*/ getDeserializedCollection(byte[] bytes) throws PersistentException {
        Collection/*<?>*/ result = null;
		try {
			ByteArrayInputStream bis = new ByteArrayInputStream(bytes);
			ObjectInputStream in = new ObjectInputStream(bis);
			result = (Collection/*<?>*/) in.readObject(); 
			in.close();
			bis.close();
		} catch (IOException e) {
            throw new PersistentException(e);
		} catch (ClassNotFoundException e) {
            throw new PersistentException(e);
		}			
	    return result;
    }

    private void modified() {
		myLastModified = System.currentTimeMillis();        
    }

    private boolean isCreated() {
    	return (myLocationFile != null);
    }
    
    private void checkState() {
		if (!isCreated()) throw new IllegalStateException("Not created "+toString());
		if (!isOpened) throw new IllegalStateException("Not opened "+toString());
    }
    
    private final class PersistentEntryImpl implements PersistentEntry {
    	private final String myKey;
    	
        protected PersistentEntryImpl(String key) {
    	    myKey = key;
    	}

    	public PersistentMap getDatabase() {
            return PersistentMapImplBase.this;
    	}
    	
    	public void addValue(String value) throws PersistentException {
    		Collection/*<String>*/ result = PersistentMapImplBase.this.getCollection(getKey());
    		if (result == null) {
    			result = new HashSet/*<String>*/();
    		}
    		result.add(value);
    		PersistentMapImplBase.this.putCollection(getKey(), result);
    	}

    	public String getKey() {
    		return myKey;
    	}

    	public String[] getValues() throws PersistentException {
    		Collection/*<String>*/ result = PersistentMapImplBase.this.getCollection(getKey());
    		return (result == null) ? (new String[0]) : ((String[])result.toArray(new String[0]));
    	}

    	public void remove() throws PersistentException {
    	    PersistentMapImplBase.this.removeCollection(getKey());
    	}

    	public String toString() {
    		String result = getKey();
    		try {
    			result += Arrays.asList(getValues());
    		} catch (PersistentException e) {
    			e.printStackTrace();
    		}
    		return result;
    	}
    	
    }    

    private final class PersistentIteratorImpl implements PersistentIterator {
        private StringIterator myIterator;
        
        public PersistentIteratorImpl(StringIterator iterator) {
            myIterator = iterator;
        }
        
        public boolean hasNext() throws PersistentException {
            return myIterator.hasNext();
        }

        public PersistentEntry next() throws PersistentException {
            PersistentEntry result = null;
            String nextKey = myIterator.next();
            if (!(nextKey == null || "".equals(nextKey))) {
                result = getEntry(nextKey);
            }
            return result;
        }
    }
    
    public interface StringIterator {
        boolean hasNext() throws PersistentException;
        String next() throws PersistentException;
    }

}
