package org.dbxml.core.filer;

/*
 *  dbXML License, Version 1.0
 *
 *  Copyright (c) 1999-2001 The dbXML Group, L.L.C.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *  notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *  notice, this list of conditions and the following disclaimer in
 *  the documentation and/or other materials provided with the
 *  distribution.
 *
 *  3. The end-user documentation included with the redistribution,
 *  if any, must include the following acknowledgment:
 *  "This product includes software developed by
 *  The dbXML Group (http://www.dbxml.com/)."
 *  Alternately, this acknowledgment may appear in the software
 *  itself, if and wherever such third-party acknowledgments normally
 *  appear.
 *
 *  4. The names "dbXML" and "The dbXML Group" must not be used to
 *  endorse or promote products derived from this software without
 *  prior written permission. For written permission, please contact
 *  info@dbxml.com.
 *
 *  5. Products derived from this software may not be called "dbXML",
 *  nor may "dbXML" appear in their name, without prior written
 *  permission of The dbXML Group.
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED.  IN NO EVENT SHALL THE DBXML GROUP OR ITS CONTRIBUTORS
 *  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 *  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
 *  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  $Id: BTree.java,v 1.5 2005/01/17 14:47:14 dbarashev Exp $
 */
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.EOFException;
import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.Arrays;
import java.util.Stack;
import java.util.logging.Level;

import org.dbxml.core.DBException;
import org.dbxml.core.data.Value;
import org.dbxml.core.indexer.IndexQuery;
import org.exist.storage.BufferStats;
import org.exist.storage.cache.Cache;
import org.exist.storage.cache.Cacheable;
import org.exist.storage.cache.LRDCache;
import org.exist.util.ByteConversion;
import org.exist.util.VariableByteInputStream;
import org.exist.util.VariableByteOutputStream;
//import org.exist.util.VariableByteCoding;

/**
 *  BTree represents a Variable Magnitude Simple-Prefix B+Tree File. A BTree is
 *  a bit flexible in that it can be used for set or map-based indexing.
 *  HashFiler uses the BTree as a set for producing RecordSet entries. The
 *  Indexers use BTree as a map for indexing entity and attribute values in
 *  Documents. <br>
 *  <br>
 *  For those who don't know how a Simple-Prefix B+Tree works, the primary
 *  distinction is that instead of promoting actual keys to branch pages, when
 *  leaves are split, a shortest-possible separator is generated at the pivot.
 *  That separator is what is promoted to the parent branch (and continuing up
 *  the list). As a result, actual keys and pointers can only be found at the
 *  leaf level. This also affords the index the ability to ignore costly merging
 *  and redistribution of pages when deletions occur. Deletions only affect leaf
 *  pages in this implementation, and so it is entirely possible for a leaf page
 *  to be completely empty after all of its keys have been removed. <br>
 *  <br>
 *  Also, the Variable Magnitude attribute means that the btree attempts to
 *  store as many values and pointers on one page as is possible.
 */

public class BTree extends Paged implements Storage {
	protected final static byte LEAF = 1;
	protected final static byte BRANCH = 2;
	protected final static byte STREAM = 3;

//	public final static long KEY_NOT_FOUND = -1;

	private BTreeFileHeader fileHeader;
	
	protected int buffers = 1024;
	protected Cache cache;

	public BTree() {
		this(256);
	}

	public BTree(int buffers) {
		super();
		this.buffers = buffers;
		fileHeader = (BTreeFileHeader) getFileHeader();
		fileHeader.setPageCount(0);
		fileHeader.setTotalCount(0);
	}

	public BTree(File file, short keyLen) {
		this();
		setFile(file);
		fileHeader.setKeyLen(keyLen);
	}

	public BTree(File file) {
		this();
		setFile(file);
	}

	public BTree(File file, int buffers) {
		this(buffers);
		setFile(file);
	}

	public boolean open() throws DBException {
		if (super.open()) {
			cache = new LRDCache(buffers);
			//cache.setFileName(getFile().getName());
			//rootNode = getBTreeNode( fileHeader.getRootPage(), null );
			return true;
		}
		return false;
	}

	public boolean create() throws DBException {
		return create((short) - 1);
	}

	public boolean create(short fixedKeyLen) throws DBException {
		if (super.create())
			try {
				cache = new LRDCache(buffers);
				//cache.setFileName(getFile().getName());
				createRootNode();
				fileHeader.setFixedKeyLen(fixedKeyLen);
				return true;
			} catch (Exception e) {
				e.printStackTrace();
			}

		return false;
	}

	public short getFixedKeyLen() {
		BTreeFileHeader fileHeader = (BTreeFileHeader) getFileHeader();
		return fileHeader.getFixedKeyLen();
	}

	public boolean drop() throws DBException {
		try {
			return getFile().delete();
		} catch (Exception e) {
			e.printStackTrace();
		}
		return true;
	}

	/**
	 *  addValue adds a Value to the BTree and associates a pointer with it. The
	 *  pointer can be used for referencing any type of data, it just so happens
	 *  that dbXML uses it for referencing pages of associated data in the BTree
	 *  file or other files.
	 *
	 *@param  value               The Value to add
	 *@param  pointer             The pointer to associate with it
	 *@return                     The previous value for the pointer (or -1)
	 *@exception  IOException     Description of the Exception
	 *@exception  BTreeException  Description of the Exception
	 */
	public long addValue(Value value, long pointer) throws IOException, BTreeException {
		return getRootNode().addValue(value, pointer);
	}

	//TODO:
	//<comment author="dbarashev">
	// if method throws exception then it doesn't need boolean return value
	// indicating "success-failure"
	//</comment>
	public boolean remove() throws DBException {
	    boolean result = false;
	    try {
			getRootNode().remove();	
			result = true;
	    } catch (IOException e) {
	        throw new DBException(e.getMessage());
	    }
		return result;
	}

	/**
	 *  removeValue removes a Value from the BTree and returns the associated
	 *  pointer for it.
	 *
	 *@param  value               The Value to remove
	 *@return                     The pointer that was associated with it
	 *@exception  IOException     Description of the Exception
	 *@exception  BTreeException  Description of the Exception
	 */
	public long removeValue(Value value) throws IOException, BTreeException {
		return getRootNode().removeValue(value);
	}

	/**
	 *  findValue finds a Value in the BTree and returns the associated pointer
	 *  for it.
	 *
	 *@param  value               The Value to find
	 *@return                     The pointer that was associated with it
	 *@exception  IOException     Description of the Exception
	 *@exception  BTreeException  Description of the Exception
	 */
	public long findValue(Value value) throws IOException, BTreeException {
		return getRootNode().findValue(value);
	}

	/**
	 *  query performs a query against the BTree and performs callback
	 *  operations to report the search results.
	 *
	 *@param  query               The IndexQuery to use (or null for everything)
	 *@param  callback            The callback instance
	 *@exception  IOException     Description of the Exception
	 *@exception  BTreeException  Description of the Exception
	 */
	public void query(IndexQuery query, BTreeCallback callback)
		throws IOException, BTreeException {
		if (query != null && query.getOperator() == IndexQuery.TRUNC_RIGHT) {
			Value val1 = query.getValue(0);
			byte data1[] = val1.getData();
			byte data2[] = new byte[data1.length];
			System.arraycopy(data1, 0, data2, 0, data1.length);
			data2[data2.length - 1] += 1;
			query = new IndexQuery(query.getOperator(), val1, new Value(data2));
		}
		getRootNode().query(query, callback);
	}

	public void remove(IndexQuery query, BTreeCallback callback)
		throws IOException, BTreeException {
		if (query != null && query.getOperator() == IndexQuery.TRUNC_RIGHT) {
			Value val1 = query.getValue(0);
			byte data1[] = val1.getData();
			byte data2[] = new byte[data1.length];
			System.arraycopy(data1, 0, data2, 0, data1.length);
			data2[data2.length - 1] += 1;
			query = new IndexQuery(query.getOperator(), val1, new Value(data2));
		}
		getRootNode().remove(query, callback);
	}

	private BTreeNode getBTreeNode(long page, BTreeNode parent) {
		try {
			BTreeNode node = (BTreeNode) cache.get(page);
			if (node == null) {
				Page p = getPage(page);
				node = new BTreeNode(p);
				node.read();
			}
			node.parent = parent;
			int increment = node.getStatus() == BRANCH ? 2 : 1;
			cache.add(node, increment);
			return node;
		} catch (Exception e) {
			System.err.println(e);
			e.printStackTrace();
			return null;
		}
	}

	private BTreeNode createBTreeNode(byte status, BTreeNode parent) {
		try {
			Page p = getFreePage();
			BTreeNode node = new BTreeNode(p);
			node.ph.setStatus(status);
			node.setValues(new Value[0]);
			node.setPointers(new long[0]);
			node.parent = parent;
			node.write();
			fileHeader.incNodePageCount();
			//cache.add(node);
			return node;
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	protected void setRootNode(BTreeNode rootNode) throws IOException {
		fileHeader.setRootPage(rootNode.page.getPageNum());
		fileHeader.write();
		cache.add(rootNode, 2);
	}

	protected long createRootNode() throws IOException {
		BTreeNode root = createBTreeNode(LEAF, null);
		setRootNode(root);
		return root.page.getPageNum();
	}

	protected BTreeNode getRootNode() {
		try {
			BTreeNode node = (BTreeNode) cache.get(fileHeader.getRootPage());
			if (node == null) {
				Page p = getPage(fileHeader.getRootPage());
				node = new BTreeNode(p);
				node.read();
			}
			cache.add(node, 2);
			return node;
		} catch (Exception e) {
			System.err.println(e);
			e.printStackTrace();
			return null;
		}
	}

	public boolean flush() throws DBException {
		cache.flush();
		super.flush();
		return true;
	}

	public boolean close() throws DBException {
		flush();
		super.close();
		return true;
	}

	protected final class BTreeNode implements Cacheable {
		
		public Page page;
		private BTreePageHeader ph;
		private Value[] values = null;
		private long[] ptrs = null;
		private BTreeNode parent = null;
		private int refCount = 0;
		private int timestamp = 0;
		protected boolean saved = true;

		public BTreeNode(Page page) {
			this.page = page;
			ph = (BTreePageHeader) page.getPageHeader();
		}

		public int getReferenceCount() {
			return refCount;
		}

		public int incReferenceCount() {
			if (refCount < Cacheable.MAX_REF)
				++refCount;
			return refCount;
		}

		public void setReferenceCount(int count) {
			refCount = count;
		}
		
		/* (non-Javadoc)
		 * @see org.exist.storage.cache.Cacheable#setTimestamp(int)
		 */
		public void setTimestamp(int timestamp) {
			this.timestamp = timestamp;
		}
		
		/* (non-Javadoc)
		 * @see org.exist.storage.cache.Cacheable#getTimestamp()
		 */
		public int getTimestamp() {
			return timestamp;
		}
		
		public void sync() {
			if(isDirty())
				try {
					write();
				} catch (IOException e) {
					LOG.log(Level.WARNING, "IO error while writing page: " + page.getPageNum(), e);
				}
		}
		
		public long getKey() {
			return page.getPageNum();
		}

		public int decReferenceCount() {
			return refCount > 0 ? --refCount : 0;
		}

		public int getStatus() {
			return ph.getStatus();
		}

		public void setValues(Value[] values) {
			this.values = values;
			this.saved = false;
			ph.setValueCount((short) values.length);
		}

		public Value[] getValues() {
			return values;
		}

		public void setPointers(long[] ptrs) {
			this.saved = false;
			this.ptrs = ptrs;
		}

//		private final int _calcDataLen() {
//			int length = 0;
//			if (ptrs != null) {
//				for (int i = 0; i < ptrs.length; i++) {
//					length += VariableByteCoding.getSize(ptrs[i]);
//				}
//			}
//			//length = ptrs.length * 8;
//			for (int i = 0; i < values.length; i++)
//				length =
//					length + values[i].getLength() + ((fileHeader.getFixedKeyLen() < 0) ? 2 : 0);
//			return length;
//		}

		private int calcDataLen() {
			int length = ptrs == null ? 0 : ptrs.length << 3;
			final int keyLen = fileHeader.getFixedKeyLen() < 0 ? 2 : 0;
			length += keyLen * values.length;
			for (int i = 0; i < values.length; i++)
				length += values[i].getLength();
			return length;
		}
		
		private boolean mustSplit() {
			return calcDataLen() > fileHeader.getWorkSize();
		}
		
		public long[] getPointers() {
			return ptrs;
		}

		public boolean isDirty() {
			return !saved;
		}

		public void __read() throws IOException {
			byte[] data = page.read();
			VariableByteInputStream is = new VariableByteInputStream(data);

			// Read in the Values
			values = new Value[ph.getValueCount()];
			short valSize = fileHeader.getFixedKeyLen();
			byte[] b;
			for (int i = 0; i < values.length; i++) {
				if (fileHeader.getFixedKeyLen() < 0)
					valSize = is.readShort();
				b = new byte[valSize];

				is.read(b, 0, valSize);
				values[i] = new Value(b);
			}

			// Read in the pointers
			ptrs = new long[ph.getPointerCount()];
			for (int i = 0; i < ph.getPointerCount(); i++)
				try {
					ptrs[i] = is.readLong();
				} catch (EOFException eof) {
					break;
				}
		}

		public void read() throws IOException {
			byte[] data = page.read();
			values = new Value[ph.getValueCount()];
			short keyLen = fileHeader.getFixedKeyLen();
			short valSize = keyLen;
			int p = 0;
			for (int i = 0; i < values.length; i++) {
				if (keyLen < 0) {
					valSize = ByteConversion.byteToShort(data, p);
					p += 2;
				}
				values[i] = new Value(data, p, valSize);
				p += valSize;
			}

			//	Read in the pointers
			final int ptrCount = ph.getPointerCount();
			ptrs = new long[ptrCount];
			for (int i = 0; i < ptrCount; i++) {
				ptrs[i] = ByteConversion.byteToLong(data, p);
				p += 8;
			}
		}

		public void _read() throws IOException {
			byte[] data = page.read();
			DataInputStream is = new DataInputStream(new ByteArrayInputStream(data));

			// Read in the Values
			values = new Value[ph.getValueCount()];
			short valSize = fileHeader.getFixedKeyLen();
			byte[] b;
			for (int i = 0; i < values.length; i++) {
				if (fileHeader.getFixedKeyLen() < 0)
					valSize = is.readShort();
				b = new byte[valSize];

				is.read(b, 0, valSize);
				values[i] = new Value(b);
			}

			// Read in the pointers
			ptrs = new long[ph.getPointerCount()];
			for (int i = 0; i < ph.getPointerCount(); i++)
				try {
					ptrs[i] = is.readLong();
				} catch (EOFException eof) {
					// stop reading if eof
					// happens sometimes - why?
					eof.printStackTrace();
					if (ph.getStatus() == LEAF)
						System.out.println("leaf: " + ph.getPointerCount() + " read: " + (i + 1));
					else
						System.out.println(
							"branch: " + ph.getPointerCount() + " values: " + ph.getValueCount());
					break;
				}
			// free page data
			//page.setData(null);
		}

		public void __write() throws IOException {
			VariableByteOutputStream os =
				new VariableByteOutputStream();

			// Write out the values
			for (int i = 0; i < values.length; i++) {
				if (fileHeader.getFixedKeyLen() < 0)
					os.writeShort(values[i].getLength());
				values[i].streamTo(os);
			}

			// write out the pointers
			for (int i = 0; i < ptrs.length; i++)
				os.writeLong(ptrs[i]);

			byte[] temp = os.toByteArray();
			writeValue(page, new Value(temp));

			saved = true;
		}

		public void _write() throws IOException {
			ByteArrayOutputStream bos = new ByteArrayOutputStream(fileHeader.getWorkSize());
			DataOutputStream os = new DataOutputStream(bos);

			// Write out the Values
			for (int i = 0; i < values.length; i++) {
				if (fileHeader.getFixedKeyLen() < 0)
					os.writeShort(values[i].getLength());
				values[i].streamTo(os);
			}

			// Write out the pointers
			for (int i = 0; i < ptrs.length; i++)
				os.writeLong(ptrs[i]);
			byte[] temp = bos.toByteArray();
			writeValue(page, new Value(temp));

			//if (fileHeader.isDirty())
			//	fileHeader.write();
			saved = true;
		}

		public void write() throws IOException {
			final byte[] temp = new byte[fileHeader.getWorkSize()];
			int p = 0;
			byte[] data;
			final int keyLen = fileHeader.getFixedKeyLen();
			for (int i = 0; i < values.length; i++) {
				if (keyLen < 0) {
					ByteConversion.shortToByte((short) values[i].getLength(), temp, p);
					p += 2;
				}
				data = values[i].getData();
				System.arraycopy(data, 0, temp, p, data.length);
				p += data.length;
			}

			for (int i = 0; i < ptrs.length; i++) {
				ByteConversion.longToByte(ptrs[i], temp, p);
				p += 8;
			}
			writeValue(page, new Value(temp));
			saved = true;
		}

		public BTreeNode getChildNode(int idx) {
			if (ph.getStatus() == BRANCH && idx >= 0 && idx < ptrs.length) {
				return getBTreeNode(ptrs[idx], this);			    
			}
			return null;
		}

		public void remove() throws IOException, BTreeException {
			if (ph.getStatus() == BRANCH)
				for (int i = 0; i < ptrs.length; i++)
					getChildNode(i).remove();

			cache.remove(this);
			unlinkPages(page);
		}

		public long removeValue(Value value) throws IOException, BTreeException {
			int idx = Arrays.binarySearch(values, value);
			switch (ph.getStatus()) {
				case BRANCH :
					if (idx < 0)
						idx = - (idx + 1);
					return getChildNode(idx).removeValue(value);
				case LEAF :
					if (idx < 0) {
						return KEY_NOT_FOUND;					    
					}
					long oldPtr = ptrs[idx];

					setValues(deleteArrayValue(values, idx));
					setPointers(deleteArrayLong(ptrs, idx));
					//cache.add(this);
					return oldPtr;

				default :
					throw new BTreeException("Invalid Page Type In removeValue");
			}
		}

		public long addValue(Value value, long pointer) throws IOException, BTreeException {
			int idx = Arrays.binarySearch(values, value);

			switch (ph.getStatus()) {
				case BRANCH :
					if (idx < 0)
						idx = - (idx + 1);
					return getChildNode(idx).addValue(value, pointer);
				case LEAF :
					if (idx >= 0) {
						// Value was found... Overwrite
						long oldPtr = ptrs[idx];
						ptrs[idx] = pointer;

						setValues(values);
						setPointers(ptrs);
						//write();
						//cache.add(this);
						return oldPtr;
					}
					// Value was not found
					idx = - (idx + 1);

					// Check to see if we've exhausted the block
					//boolean split = ph.getDataLen() + 6 + value.getLength() > fileHeader.getWorkSize();
					setValues(insertArrayValue(values, value.trim(),
					//new Value(value.getData()),
					idx));
					setPointers(insertArrayLong(ptrs, pointer, idx));
					//cache.add(this);
					boolean split = calcDataLen() > fileHeader.getWorkSize();
					if (split)
						split();
					
					//write();
					return -1;
				default :
					throw new BTreeException("Invalid Page Type In addValue");
			}
		}

		public void promoteValue(Value value, long rightPointer)
			throws IOException, BTreeException {
			// Check to see if we've exhausted the block
			//boolean split = ph.getDataLen() + 6 + value.getLength() > fileHeader.getWorkSize();
			int idx = Arrays.binarySearch(values, value);
			if (idx < 0)
				idx = - (idx + 1);

			setValues(insertArrayValue(values, value, idx));
			setPointers(insertArrayLong(ptrs, rightPointer, idx + 1));
			cache.add(this); // this is a very-very important line!!! Never comment it!
			boolean split = calcDataLen() > fileHeader.getWorkSize();
			if (split)
				split();
			//write();
		}

		public Value getSeparator(Value value1, Value value2) {
			int idx = value1.compareTo(value2);
			byte[] b = new byte[Math.abs(idx)];
			System.arraycopy(value2.getData(), 0, b, 0, b.length);
			return new Value(b);
		}

		public void split() throws IOException, BTreeException {
			Value[] leftVals;
			Value[] rightVals;
			long[] leftPtrs;
			long[] rightPtrs;
			Value separator;

			final short vc = ph.getValueCount();
			final int pivot = vc / 2;
			// Split the node into two nodes
			switch (ph.getStatus()) {
				case BRANCH :
					leftVals = new Value[pivot];
					leftPtrs = new long[leftVals.length + 1];
					rightVals = new Value[vc - (pivot + 1)];
					rightPtrs = new long[rightVals.length + 1];

					System.arraycopy(values, 0, leftVals, 0, leftVals.length);
					System.arraycopy(ptrs, 0, leftPtrs, 0, leftPtrs.length);
					System.arraycopy(values, leftVals.length + 1, rightVals, 0, rightVals.length);
					System.arraycopy(ptrs, leftPtrs.length, rightPtrs, 0, rightPtrs.length);

					separator = values[leftVals.length];
					break;
				case LEAF :
					leftVals = new Value[pivot];
					leftPtrs = new long[leftVals.length];
					rightVals = new Value[vc - pivot];
					rightPtrs = new long[rightVals.length];

					System.arraycopy(values, 0, leftVals, 0, leftVals.length);
					System.arraycopy(ptrs, 0, leftPtrs, 0, leftPtrs.length);
					System.arraycopy(values, leftVals.length, rightVals, 0, rightVals.length);
					System.arraycopy(ptrs, leftPtrs.length, rightPtrs, 0, rightPtrs.length);

					separator = leftVals[leftVals.length - 1];
					break;
				default :
					throw new BTreeException("Invalid Page Type In split");
			}

			setValues(leftVals);
			setPointers(leftPtrs);
			//cache.add(this);

			// Promote the pivot to the parent branch
			if (parent == null) {
				// This can only happen if this is the root
				parent = createBTreeNode(BRANCH, null);

				final BTreeNode rNode = createBTreeNode(ph.getStatus(), parent);
				rNode.setValues(rightVals);
				rNode.setPointers(rightPtrs);
				cache.add(rNode); // this is a very-very important line!!! Never comment it!

				parent.setValues(new Value[] { separator });
				parent.setPointers(new long[] { page.getPageNum(), rNode.page.getPageNum()});
				cache.add(parent);
				setRootNode(parent);
				if(rNode.mustSplit()) {
					LOG.fine(getFile().getName() + " right node requires second split: " + rNode.calcDataLen());
					rNode.split();
				}
			} else {
				final BTreeNode rNode = createBTreeNode(ph.getStatus(), parent);
				rNode.setValues(rightVals);
				rNode.setPointers(rightPtrs);
				cache.add(rNode); // this is a very-very important line!!! Never comment it!
				parent.promoteValue(separator, rNode.page.getPageNum());
				if(rNode.mustSplit()) {
					LOG.fine(getFile().getName() + " right node requires second split: " + rNode.calcDataLen());
					rNode.split();
				}
			}
			if(mustSplit()) {
				LOG.fine(getFile().getName() + "left node requires second split: " + calcDataLen());
				split();
			}
		}

		/////////////////////////////////////////////////////////////////

		public long findValue(Value value) throws IOException, BTreeException {
			int idx = Arrays.binarySearch(values, value);
			switch (ph.getStatus()) {
				case BRANCH :
					if (idx < 0)
						idx = - (idx + 1);
					BTreeNode child = getChildNode(idx);
					if (child == null)
						throw new BTreeException(
							"unexpected "
								+ ptrs[idx]
								+ ", "
								+ page.getPageNum()
								+ ": value '"
								+ value.toString()
								+ "' doesn't exist");
					return child.findValue(value);
				case LEAF :
					if (idx < 0) {
						return KEY_NOT_FOUND;
						//throw new BTreeException("Value doesn't exist");
					}
					return ptrs[idx];
				default :
					throw new BTreeException("Invalid Page Type In findValue");
			}
		}

		// query is a BEAST of a method
		// --- added support for right truncated search terms (wolf)
		public void query(IndexQuery query, BTreeCallback callback)
			throws IOException, BTreeException {
			if (query != null
				&& query.getOperator() != IndexQuery.ANY
				&& query.getOperator() != IndexQuery.TRUNC_LEFT) {
				Value[] qvals = query.getValues();
				int leftIdx = Arrays.binarySearch(values, qvals[0]);
				int rightIdx =
					qvals.length > 1
						? Arrays.binarySearch(values, qvals[qvals.length - 1])
						: leftIdx;
				boolean pos = query.getOperator() >= 0;
				switch (ph.getStatus()) {

					case BRANCH :
						if (leftIdx < 0)
							leftIdx = - (leftIdx + 1);
						if (rightIdx < 0)
							rightIdx = - (rightIdx + 1);

						switch (query.getOperator()) {
							case IndexQuery.BWX :
							case IndexQuery.NBWX :
							case IndexQuery.BW :
							case IndexQuery.NBW :
							case IndexQuery.IN :
							case IndexQuery.NIN :
							case IndexQuery.TRUNC_RIGHT :
								for (int i = 0; i < ptrs.length; i++)
									if ((i >= leftIdx && i <= rightIdx) == pos)
										getChildNode(i).query(query, callback);
								break;
							case IndexQuery.EQ :
							case IndexQuery.NEQ :
								for (int i = 0; i < ptrs.length; i++)
									if (!pos || i == leftIdx)
										getChildNode(i).query(query, callback);

							case IndexQuery.LT :
							case IndexQuery.GEQ :
								for (int i = 0; i < ptrs.length; i++)
									if ((pos && (i <= leftIdx)) || (!pos && (i >= leftIdx)))
										getChildNode(i).query(query, callback);
								break;
							case IndexQuery.GT :
							case IndexQuery.LEQ :
								for (int i = 0; i < ptrs.length; i++)
									if ((pos && (i >= leftIdx)) || (!pos && (i <= leftIdx)))
										getChildNode(i).query(query, callback);
								break;
							default :
								// If it's not implemented, we walk the tree

								for (int i = 0; i < ptrs.length; i++)
									getChildNode(i).query(query, callback);
								break;
						}
						break;
					case LEAF :
						switch (query.getOperator()) {
							case IndexQuery.EQ :
								if (leftIdx >= 0)
									callback.indexInfo(values[leftIdx], ptrs[leftIdx]);
								break;
							case IndexQuery.NEQ :
								for (int i = 0; i < ptrs.length; i++)
									if (i != leftIdx)
										callback.indexInfo(values[i], ptrs[i]);
								break;
							case IndexQuery.BWX :
							case IndexQuery.NBWX :
							case IndexQuery.BW :
							case IndexQuery.NBW :
								if (leftIdx < 0)
									leftIdx = - (leftIdx + 1);
								if (rightIdx < 0)
									rightIdx = - (rightIdx + 1);
								for (int i = 0; i < ptrs.length; i++)
									if ((pos && (i >= leftIdx && i <= rightIdx))
										|| (!pos && (i <= leftIdx || i >= rightIdx))) {
										if (query.testValue(values[i]))
											callback.indexInfo(values[i], ptrs[i]);
									}
								break;
							case IndexQuery.TRUNC_RIGHT :
								if (leftIdx < 0)
									leftIdx = - (leftIdx + 1);
								if (rightIdx < 0)
									rightIdx = - (rightIdx + 1);
								for (int i = 0; i < ptrs.length; i++)
									if ((pos && (i >= leftIdx && i < rightIdx))
										|| (!pos && (i <= leftIdx || i >= rightIdx))) {
										if (query.testValue(values[i]))
											callback.indexInfo(values[i], ptrs[i]);
									}
								break;
							case IndexQuery.IN :
							case IndexQuery.NIN :
								if (leftIdx < 0)
									leftIdx = - (leftIdx + 1);
								if (rightIdx < 0)
									rightIdx = - (rightIdx + 1);
								for (int i = 0; i < ptrs.length; i++)
									if (!pos || (i >= leftIdx && i <= rightIdx))
										if (query.testValue(values[i]))
											callback.indexInfo(values[i], ptrs[i]);
								break;
							case IndexQuery.LT :
							case IndexQuery.GEQ :
								if (leftIdx < 0)
									leftIdx = - (leftIdx + 1);
								for (int i = 0; i < ptrs.length; i++)
									if ((pos && (i <= leftIdx)) || (!pos && (i >= leftIdx)))
										if (query.testValue(values[i]))
											callback.indexInfo(values[i], ptrs[i]);
								break;
							case IndexQuery.GT :
							case IndexQuery.LEQ :
								if (leftIdx < 0)
									leftIdx = - (leftIdx + 1);
								for (int i = 0; i < ptrs.length; i++)
									if ((pos && (i >= leftIdx)) || (!pos && (i <= leftIdx))) {
										if (query.testValue(values[i]))
											callback.indexInfo(values[i], ptrs[i]);
										else if (query.getOperator() == IndexQuery.TRUNC_RIGHT)
											break;
									}
								break;
							default :
								// If it's not implemented, it falls right through

								for (int i = 0; i < ptrs.length; i++)
									if (query.testValue(values[i]))
										callback.indexInfo(values[i], ptrs[i]);
								break;
						}
						break;
					default :
						throw new BTreeException("Invalid Page Type In query");
				}

			} else
				// No Query - Just Walk The Tree
				switch (ph.getStatus()) {
					case BRANCH :
						for (int i = 0; i < ptrs.length; i++)
							getChildNode(i).query(query, callback);
						break;
					case LEAF :
						for (int i = 0; i < values.length; i++)
							if (query.getOperator() != IndexQuery.TRUNC_LEFT
								|| query.testValue(values[i]))
								callback.indexInfo(values[i], ptrs[i]);
						break;
					default :
						throw new BTreeException("Invalid Page Type In query");
				}

		}
		
		// query is a BEAST of a method
		// --- added support for right truncated search terms (wolf)
		public void remove(IndexQuery query, BTreeCallback callback)
			throws IOException, BTreeException {
			if (query != null
				&& query.getOperator() != IndexQuery.ANY
				&& query.getOperator() != IndexQuery.TRUNC_LEFT) {
				Value[] qvals = query.getValues();
				int leftIdx = Arrays.binarySearch(values, qvals[0]);
				int rightIdx =
					qvals.length > 1
						? Arrays.binarySearch(values, qvals[qvals.length - 1])
						: leftIdx;
				boolean pos = query.getOperator() >= 0;
				switch (ph.getStatus()) {

					case BRANCH :
						if (leftIdx < 0)
							leftIdx = - (leftIdx + 1);
						if (rightIdx < 0)
							rightIdx = - (rightIdx + 1);

						switch (query.getOperator()) {
							case IndexQuery.BWX :
							case IndexQuery.NBWX :
							case IndexQuery.BW :
							case IndexQuery.NBW :
							case IndexQuery.IN :
							case IndexQuery.NIN :
							case IndexQuery.TRUNC_RIGHT :
								for (int i = 0; i < ptrs.length; i++)
									if ((i >= leftIdx && i <= rightIdx) == pos)
										getChildNode(i).remove(query, callback);
								break;
							case IndexQuery.EQ :
							case IndexQuery.NEQ :
								for (int i = 0; i < ptrs.length; i++)
									if (!pos || i == leftIdx)
										getChildNode(i).remove(query, callback);

							case IndexQuery.LT :
							case IndexQuery.GEQ :
								for (int i = 0; i < ptrs.length; i++)
									if ((pos && (i <= leftIdx)) || (!pos && (i >= leftIdx)))
										getChildNode(i).remove(query, callback);
								break;
							case IndexQuery.GT :
							case IndexQuery.LEQ :
								for (int i = 0; i < ptrs.length; i++)
									if ((pos && (i >= leftIdx)) || (!pos && (i <= leftIdx)))
										getChildNode(i).remove(query, callback);
								break;
							default :
								// If it's not implemented, we walk the tree

								for (int i = 0; i < ptrs.length; i++)
									getChildNode(i).remove(query, callback);
								break;
						}
						break;
					case LEAF :
						switch (query.getOperator()) {
							case IndexQuery.EQ :
								if (leftIdx >= 0) {
									if (callback != null)
										callback.indexInfo(values[leftIdx], ptrs[leftIdx]);
									setValues(deleteArrayValue(values, leftIdx));
									setPointers(deleteArrayLong(ptrs, leftIdx));
									//cache.add(this);
								}
								break;
							case IndexQuery.NEQ :
								for (int i = 0; i < ptrs.length; i++)
									if (i != leftIdx) {
										if (callback != null)
											callback.indexInfo(values[i], ptrs[i]);
										setValues(deleteArrayValue(values, i));
										setPointers(deleteArrayLong(ptrs, i));
										//cache.add(this);
									}
								break;
							case IndexQuery.BWX :
							case IndexQuery.NBWX :
							case IndexQuery.BW :
							case IndexQuery.NBW :
								if (leftIdx < 0)
									leftIdx = - (leftIdx + 1);
								if (rightIdx < 0)
									rightIdx = - (rightIdx + 1);
								for (int i = 0; i < ptrs.length; i++)
									if ((pos && (i >= leftIdx && i <= rightIdx))
										|| (!pos && (i <= leftIdx || i >= rightIdx))) {
										if (query.testValue(values[i])) {
											if (callback != null)
												callback.indexInfo(values[i], ptrs[i]);
											setValues(deleteArrayValue(values, i));
											setPointers(deleteArrayLong(ptrs, i));
											--i;
											//cache.add(this);
										}
									}
								break;
							case IndexQuery.TRUNC_RIGHT :
								if (leftIdx < 0)
									leftIdx = - (leftIdx + 1);
								if (rightIdx < 0)
									rightIdx = - (rightIdx + 1);
								for (int i = 0; i < ptrs.length; i++)
									if ((pos && (i >= leftIdx && i < rightIdx))
										|| (!pos && (i <= leftIdx || i >= rightIdx))) {
										if (query.testValue(values[i])) {
											if (callback != null)
												callback.indexInfo(values[i], ptrs[i]);
											setValues(deleteArrayValue(values, i));
											setPointers(deleteArrayLong(ptrs, i));
											--i;
											//cache.add(this);
										}
									}
								break;
							case IndexQuery.IN :
							case IndexQuery.NIN :
								if (leftIdx < 0)
									leftIdx = - (leftIdx + 1);
								if (rightIdx < 0)
									rightIdx = - (rightIdx + 1);
								for (int i = 0; i < ptrs.length; i++)
									if (!pos || (i >= leftIdx && i <= rightIdx))
										if (query.testValue(values[i])) {
											if (callback != null)
												callback.indexInfo(values[i], ptrs[i]);
											setValues(deleteArrayValue(values, i));
											setPointers(deleteArrayLong(ptrs, i));
											--i;
											//cache.add(this);
										}
								break;
							case IndexQuery.LT :
							case IndexQuery.GEQ :
								if (leftIdx < 0)
									leftIdx = - (leftIdx + 1);
								for (int i = 0; i < ptrs.length; i++)
									if ((pos && (i <= leftIdx)) || (!pos && (i >= leftIdx)))
										if (query.testValue(values[i])) {
											if (callback != null)
												callback.indexInfo(values[i], ptrs[i]);
											setValues(deleteArrayValue(values, i));
											setPointers(deleteArrayLong(ptrs, i));
											--i;
											//cache.add(this);
										}
								break;
							case IndexQuery.GT :
							case IndexQuery.LEQ :
								if (leftIdx < 0)
									leftIdx = - (leftIdx + 1);
								for (int i = 0; i < ptrs.length; i++)
									if ((pos && (i >= leftIdx)) || (!pos && (i <= leftIdx))) {
										if (query.testValue(values[i])) {
											if (callback != null)
												callback.indexInfo(values[i], ptrs[i]);
											setValues(deleteArrayValue(values, i));
											setPointers(deleteArrayLong(ptrs, i));
											--i;
											//cache.add(this);
										} else if (query.getOperator() == IndexQuery.TRUNC_RIGHT)
											break;
									}
								break;
							default :
								// If it's not implemented, it falls right through

								for (int i = 0; i < ptrs.length; i++)
									if (query.testValue(values[i])) {
										if (callback != null)
											callback.indexInfo(values[i], ptrs[i]);
										setValues(deleteArrayValue(values, i));
										setPointers(deleteArrayLong(ptrs, i));
										--i;
										//cache.add(this);
									}

								break;
						}
						break;
					default :
						throw new BTreeException("Invalid Page Type In query");
				}

			} else
				// No Query - Just Walk The Tree
				switch (ph.getStatus()) {
					case BRANCH :
						for (int i = 0; i < ptrs.length; i++) {
							if (callback != null)
								callback.indexInfo(values[i], ptrs[i]);
							setValues(deleteArrayValue(values, i));
							setPointers(deleteArrayLong(ptrs, i));
							--i;
							//cache.add(this);
						}
						break;
					case LEAF :
						for (int i = 0; i < values.length; i++)
							if (query.getOperator() != IndexQuery.TRUNC_LEFT
								|| query.testValue(values[i])) {
								if (callback != null)
									callback.indexInfo(values[i], ptrs[i]);
								setValues(deleteArrayValue(values, i));
								setPointers(deleteArrayLong(ptrs, i));
								--i;
								//cache.add(this);
							}
						break;
					default :
						throw new BTreeException("Invalid Page Type In query");
				}

		}
		
		/* (non-Javadoc)
		 * @see org.exist.storage.cache.Cacheable#allowUnload()
		 */
		public boolean allowUnload() {
			return true;
		}
	}

	////////////////////////////////////////////////////////////////////

    protected class BTreeIterator implements ValuesIterator {
        private Stack myIterationPositionStack = new Stack();
        private BTreeNode myCurrentNode;

        protected BTreeIterator(BTreeNode startNode) {
            if (!(startNode.getPointers().length == 0)) {
                myCurrentNode = startNode;
                myIterationPositionStack.push(new Integer(0));                
            }
        }

		public boolean hasNext() throws BTreeException {
            return !myIterationPositionStack.empty();
        }
		
		public Value next() throws BTreeException {
            Value result = null;
            if (hasNext()) {
                int currentIterationPosition = ((Integer) myIterationPositionStack.peek()).intValue();
                if (currentIterationPosition < myCurrentNode.getPointers().length) {
                    if (myCurrentNode.getStatus() == LEAF) {
                        result = myCurrentNode.getValues()[currentIterationPosition];
                        increaseIterationPosition(currentIterationPosition);
                    } else if (myCurrentNode.getStatus() == BRANCH) {
                        myCurrentNode = myCurrentNode.getChildNode(currentIterationPosition);
                        myIterationPositionStack.push(new Integer(0));
                        result = next();
                    }
                }
            }
            return result;
        }
		
		private void increaseIterationPosition(int currentIterationPosition) throws BTreeException {
            currentIterationPosition++;
            if (currentIterationPosition == myCurrentNode.getPointers().length) {
                myIterationPositionStack.pop(); // remove finished node
                myCurrentNode = myCurrentNode.parent;
                if (hasNext()) { 
                    increaseIterationPosition(((Integer) myIterationPositionStack.peek()).intValue());                    
                }
            } else if (currentIterationPosition < myCurrentNode.getPointers().length) {
                myIterationPositionStack.pop();
                myIterationPositionStack.push(new Integer(currentIterationPosition));
            }
		}
	}
		
	public ValuesIterator iterator() {
	    return new BTreeIterator(getRootNode());
	}

	public FileHeader createFileHeader() {
		return new BTreeFileHeader(PAGE_SIZE);
	}

	public FileHeader createFileHeader(boolean read) throws IOException {
		return new BTreeFileHeader(read);
	}

	public FileHeader createFileHeader(long pageCount) {
		return new BTreeFileHeader(pageCount, PAGE_SIZE);
	}

	public FileHeader createFileHeader(long pageCount, int pageSize) {
		return new BTreeFileHeader(pageCount, pageSize);
	}

	public PageHeader createPageHeader() {
		return new BTreePageHeader();
	}

	public BufferStats getIndexBufferStats() {
		return new BufferStats(
			cache.getBuffers(),
			cache.getUsedBuffers(),
			cache.getHits(),
			cache.getFails());
	}

	public void printStatistics() {
		StringBuffer buf = new StringBuffer();
		buf.append(getFile().getName()).append(" INDEX ");
		buf.append(cache.getBuffers()).append(" / ");
		buf.append(cache.getUsedBuffers()).append(" / ");
		buf.append(cache.getHits()).append(" / ");
		buf.append(cache.getFails());
		LOG.info(buf.toString());
	}

	protected class BTreeFileHeader extends FileHeader {
		private long rootPage = 0;
		private short keyLen = -1;
		private int nodePageCount = 0;
		private short fixedLen = -1;

		public BTreeFileHeader() {
			super();
		}

		public BTreeFileHeader(long pageCount) {
			super(pageCount);
		}

		public BTreeFileHeader(long pageCount, int pageSize) {
			super(pageCount, pageSize);
		}

//		public BTreeFileHeader(long pageCount, int pageSize, byte blockSize) {
//			super(pageCount, pageSize, blockSize);
//		}

		public BTreeFileHeader(int pageSize) {
			super(pageSize);
		}

		public BTreeFileHeader(boolean read) throws IOException {
			super(read);
		}

		public void read(RandomAccessFile raf) throws IOException {
			super.read(raf);
			rootPage = raf.readLong();
			keyLen = raf.readShort();
			fixedLen = raf.readShort();
			nodePageCount = raf.readInt();
		}

		public void write(RandomAccessFile raf) throws IOException {
			super.write(raf);
			raf.writeLong(rootPage);
			raf.writeShort(keyLen);
			raf.writeShort(fixedLen);
			raf.writeInt(nodePageCount);
		}

		/**
		 *  The root page of the storage tree
		 *
		 *@param  rootPage  The new rootPage value
		 */
		public final void setRootPage(long rootPage) {
			this.rootPage = rootPage;
			setDirty(true);
		}

		/**
		 *  The root page of the storage tree
		 *
		 *@return    The rootPage value
		 */
		public final long getRootPage() {
			return rootPage;
		}

		public short getKeyLen() {
			return keyLen;
		}

		public void setKeyLen(short keyLen) {
			this.keyLen = keyLen;
		}

		public short getFixedKeyLen() {
			return fixedLen;
		}

		public void setFixedKeyLen(short keyLen) {
			this.fixedLen = keyLen;
		}

		public void incNodePageCount() {
			++nodePageCount;
		}

		public int getNodePageCount() {
			return nodePageCount;
		}
	}

	protected class BTreePageHeader extends PageHeader {
		private short valueCount = 0;

		public BTreePageHeader() {
			super();
		}

		public BTreePageHeader(byte[] data, int offset) throws IOException {
			super(data, offset);
		}

		public int read(byte[] data, int offset) throws IOException {
			offset = super.read(data, offset);
			//if (getStatus() == UNUSED)
			//	return;
			valueCount = ByteConversion.byteToShort(data, offset);
			return offset + 2;
		}

		public int write(byte[] data, int offset) throws IOException {
			offset = super.write(data, offset);
			ByteConversion.shortToByte(valueCount, data, offset);
			return offset + 2;
		}

		public final void setValueCount(short valueCount) {
			this.valueCount = valueCount;
			setDirty(true);
		}

		/**
		 *  The number of values stored by this page
		 *
		 *@return    The valueCount value
		 */
		public final short getValueCount() {
			return valueCount;
		}

		/**
		 *  The number of pointers stored by this page
		 *
		 *@return    The pointerCount value
		 */
		public final short getPointerCount() {
			if (getStatus() == BRANCH) {
				return (short) (valueCount + 1);			    
			}
			return valueCount;
		}
	}
}
