package org.dbxml.core.filer;

/*
 *  dbXML License, Version 1.0
 *
 *
 *  Copyright (c) 1999-2001 The dbXML Group, L.L.C.
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *  1. Redistributions of source code must retain the above copyright
 *  notice, this list of conditions and the following disclaimer.
 *
 *  2. Redistributions in binary form must reproduce the above copyright
 *  notice, this list of conditions and the following disclaimer in
 *  the documentation and/or other materials provided with the
 *  distribution.
 *
 *  3. The end-user documentation included with the redistribution,
 *  if any, must include the following acknowledgment:
 *  "This product includes software developed by
 *  The dbXML Group (http://www.dbxml.com/)."
 *  Alternately, this acknowledgment may appear in the software
 *  itself, if and wherever such third-party acknowledgments normally
 *  appear.
 *
 *  4. The names "dbXML" and "The dbXML Group" must not be used to
 *  endorse or promote products derived from this software without
 *  prior written permission. For written permission, please contact
 *  info@dbxml.com.
 *
 *  5. Products derived from this software may not be called "dbXML",
 *  nor may "dbXML" appear in their name, without prior written
 *  permission of The dbXML Group.
 *
 *  THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 *  WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 *  OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 *  DISCLAIMED.  IN NO EVENT SHALL THE DBXML GROUP OR ITS CONTRIBUTORS
 *  BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 *  OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
 *  OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
 *  BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */
import java.io.File;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.Arrays;

//import org.apache.log4j.Logger;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.dbxml.core.DBException;
import org.dbxml.core.data.Value;
import org.exist.util.ByteConversion;

/**
 *  Paged is a paged file foundation that is used by both the BTree class and
 *  the HashFiler. It provides flexible paged I/O and page caching
 *  functionality.
 *
 */

public abstract class Paged {
	
	protected final static Logger LOG = Logger.getLogger(Paged.class.getName());
	
	protected final static byte UNUSED = 0;	
	public static int PAGE_SIZE = 4096;
	
	private RandomAccessFile raf;
	private File file;
	private FileHeader fileHeader;
	private boolean readOnly = false;
	private boolean fileIsNew = false;
	
	private byte[] tempPageData = null;
	private byte[] tempHeaderData = null;

//	private long lastPage = -1;
	
	public Paged() {
		fileHeader = createFileHeader();
		tempPageData = new byte[fileHeader.pageSize];
		tempHeaderData = new byte[fileHeader.pageHeaderSize];
	}

	public Paged(File file) {
		this();
		setFile(file);
	}

	public final static void setPageSize(int pageSize) {
		PAGE_SIZE = pageSize;
	}

	public final static int getPageSize() {
		return PAGE_SIZE;
	}

	public static int[] deleteArrayInt(int[] vals, int idx) {
		int[] newVals = new int[vals.length - 1];
		if (idx > 0)
			System.arraycopy(vals, 0, newVals, 0, idx);
		if (idx < newVals.length)
			System.arraycopy(vals, idx + 1, newVals, idx, newVals.length - idx);
		return newVals;
	}

	public static long[] deleteArrayLong(long[] vals, int idx) {
		long[] newVals = new long[vals.length - 1];
		if (idx > 0)
			System.arraycopy(vals, 0, newVals, 0, idx);
		if (idx < newVals.length)
			System.arraycopy(vals, idx + 1, newVals, idx, newVals.length - idx);
		return newVals;
	}

	public static short[] deleteArrayShort(short[] vals, int idx) {
		short[] newVals = new short[vals.length - 1];
		if (idx > 0)
			System.arraycopy(vals, 0, newVals, 0, idx);
		if (idx < newVals.length)
			System.arraycopy(vals, idx + 1, newVals, idx, newVals.length - idx);
		return newVals;
	}

	public static Value[] deleteArrayValue(Value[] vals, int idx) {
		Value[] newVals = new Value[vals.length - 1];
		if (idx > 0)
			System.arraycopy(vals, 0, newVals, 0, idx);
		if (idx < newVals.length)
			System.arraycopy(vals, idx + 1, newVals, idx, newVals.length - idx);
		return newVals;
	}

	public static int[] insertArrayInt(int[] vals, int val, int idx) {
		int[] newVals = new int[vals.length + 1];
		if (idx > 0)
			System.arraycopy(vals, 0, newVals, 0, idx);
		newVals[idx] = val;
		if (idx < vals.length)
			System.arraycopy(vals, idx, newVals, idx + 1, vals.length - idx);
		return newVals;
	}

	public static long[] insertArrayLong(long[] vals, long val, int idx) {
		long[] newVals = new long[vals.length + 1];
		if (idx > 0)
			System.arraycopy(vals, 0, newVals, 0, idx);
		newVals[idx] = val;
		if (idx < vals.length)
			System.arraycopy(vals, idx, newVals, idx + 1, vals.length - idx);
		return newVals;
	}

	public static short[] insertArrayShort(short[] vals, short val, int idx) {
		short[] newVals = new short[vals.length + 1];
		if (idx > 0)
			System.arraycopy(vals, 0, newVals, 0, idx);
		newVals[idx] = val;
		if (idx < vals.length)
			System.arraycopy(vals, idx, newVals, idx + 1, vals.length - idx);
		return newVals;
	}

	// These are a bunch of utility methods for subclasses

	public static Value[] insertArrayValue(Value[] vals, Value val, int idx) {
		Value[] newVals = new Value[vals.length + 1];
		if (idx > 0)
			System.arraycopy(vals, 0, newVals, 0, idx);
		newVals[idx] = val;
		if (idx < vals.length)
			System.arraycopy(vals, idx, newVals, idx + 1, vals.length - idx);
		return newVals;
	}

	public final boolean isReadOnly() {
		return readOnly;
	}

	public boolean close() throws DBException {
	    try {
//	        flush();
            raf.close();
        } catch (IOException e) {
    		throw new DBException(e.getMessage());
        }
		return true;
	}

	public boolean create() throws DBException {
		try {
			fileHeader.write();
			return true;
		} catch (Exception e) {
			e.printStackTrace();
			throw new DBException(0, "Error creating " + file.getName());
		}
	}

	/**
	 *  createFileHeader must be implemented by a Paged implementation in order
	 *  to create an appropriate subclass instance of a FileHeader.
	 *
	 *@return    a new FileHeader
	 */
	public abstract FileHeader createFileHeader();

	/**
	 *  createFileHeader must be implemented by a Paged implementation in order
	 *  to create an appropriate subclass instance of a FileHeader.
	 *
	 *@param  read          If true, reads the FileHeader from disk
	 *@return               a new FileHeader
	 *@throws  IOException  if an exception occurs
	 */
	public abstract FileHeader createFileHeader(boolean read) throws IOException;

	/**
	 *  createFileHeader must be implemented by a Paged implementation in order
	 *  to create an appropriate subclass instance of a FileHeader.
	 *
	 *@param  pageCount  The number of pages to allocate for primary storage
	 *@return            a new FileHeader
	 */
	public abstract FileHeader createFileHeader(long pageCount);

	/**
	 *  createFileHeader must be implemented by a Paged implementation in order
	 *  to create an appropriate subclass instance of a FileHeader.
	 *
	 *@param  pageCount  The number of pages to allocate for primary storage
	 *@param  pageSize   The size of a Page (should be a multiple of a FS block)
	 *@return            a new FileHeader
	 */
	public abstract FileHeader createFileHeader(long pageCount, int pageSize);

	/**
	 *  createPageHeader must be implemented by a Paged implementation in order
	 *  to create an appropriate subclass instance of a PageHeader.
	 *
	 *@return    a new PageHeader
	 */
	public abstract PageHeader createPageHeader();

	public boolean drop() throws DBException {
	    close();
	    getFile().delete();
		return true;
	}

	public boolean exists() {
		return !fileIsNew;
	}

	public boolean flush() throws DBException {
		try {
			fileHeader.write();
		} catch (IOException ioe) {
			ioe.printStackTrace();
		    throw new DBException(ioe.getMessage());
		}
		return true;
	}

	/**
	 *  getFile returns the file object for this Paged.
	 *
	 *@return    The File
	 */
	public final File getFile() {
		return file;
	}
	
	/**
	 *  getFileHeader returns the FileHeader
	 *
	 *@return    The FileHeader
	 */
	public FileHeader getFileHeader() {
		return fileHeader;
	}

	/**
	 *  getFreePage returns the first free Page from secondary storage. If no
	 *  Pages are available, the file is grown as appropriate.
	 *
	 *@return               The next free Page
	 *@throws  IOException  if an Exception occurs
	 */
	public final Page getFreePage() throws IOException {
		Page p = null;
		synchronized (fileHeader) {
			long pageNum = fileHeader.firstFreePage;
			if (pageNum != -1) {
				// Steal a deleted page
				p = new Page(pageNum);
				p.read();
//				if(getFile().getName().equals("words.dbx"))
//					System.out.println("reusing page " + pageNum + "; next = " + p.header.nextPage);
				fileHeader.firstFreePage = p.header.nextPage;
				if (fileHeader.firstFreePage == -1)
					fileHeader.setLastFreePage(-1);
			} else {
				// Grow the file
				pageNum = fileHeader.totalCount;
//				if(getFile().getName().equals("words.dbx"))
//					System.out.println("growing the file: " + fileHeader.totalCount);
				if(pageNum == Integer.MAX_VALUE) {
					throw new IOException("page limit reached: " + pageNum);
				}
				fileHeader.setTotalCount(pageNum + 1);
				p = new Page(pageNum);
				p.read();
			}
//			if(getFile().getName().equals("words.dbx"))
//				printFreeSpaceList();
		}
		// Initialize The Page Header (Cleanly)
		p.header.setNextPage(-1);
		p.header.setStatus(UNUSED);
		fileHeader.setDirty(true);
		return p;
	}

	/**
	 *  getPage returns the page specified by pageNum.
	 *
	 *@param  pageNum       The Page number
	 *@return               The requested Page
	 */
	protected final Page getPage(long pageNum) {
		final Page p = new Page(pageNum);
		return p;
	}

	/**
	 *  Gets the opened attribute of the Paged object
	 *
	 *@return    The opened value
	 */
	public boolean isOpened() {
		return true;
	}

	/**
	 *  Description of the Method
	 *
	 *@return                  Description of the Return Value
	 *@exception  DBException  Description of the Exception
	 */
	public boolean open() throws DBException {
		try {
			if (exists()) {
				fileHeader.read();
				return true;
			}
			return false;
		} catch (Exception e) {
			e.printStackTrace();
			throw new DBException(0, "Error opening " + file.getName());
		}
	}

//	/**
//	 *  Debug
//	 *
//	 *@exception  IOException  Description of the Exception
//	 */
//	private void printFreeSpaceList() throws IOException {
//		long pageNum = fileHeader.firstFreePage;
//		System.out.println("first free page: " + pageNum);
//		Page next;
//		System.out.println("free pages for " + getFile().getName());
//		while (pageNum != -1) {
//			next = getPage(pageNum);
//			next.read();
//			System.out.print(pageNum + ";");
//			pageNum = next.header.nextPage;
//		}
//		System.out.println();
//	}

	/**
	 *  setFile sets the file object for this Paged.
	 *
	 *@param  file  The File
	 */
	protected final void setFile(final File file) {
		this.file = file;
		fileIsNew = !file.exists();
		try {
			if ((!file.exists()) || file.canWrite()) {
				raf = new RandomAccessFile(file, "rw");
			} else {
				readOnly = true;
				raf = new RandomAccessFile(file, "r");
			}
		} catch (Exception e) {
		}
	}

	/**
	 *  unlinkPages unlinks a set of pages starting at the specified Page.
	 *
	 *@param  page          The starting Page to unlink
	 *@throws  IOException  if an Exception occurs
	 */
	public void unlinkPages(Page page) throws IOException {
		if (page != null) {
//			if(getFile().getName().equals("words.dbx"))
//				LOG.debug("freeing page " + page.getPageNum());
			// Walk the chain and add it to the unused list
//			long firstPage = page.pageNum;
			page.header.setStatus(UNUSED);
			synchronized (fileHeader) {
				if (fileHeader.firstFreePage == -1) {
					fileHeader.setFirstFreePage(page.pageNum);
					page.header.setNextPage(-1);
				} else {
					Page p = getPage(fileHeader.firstFreePage);
					fileHeader.setFirstFreePage(page.pageNum);
					page.header.setNextPage(p.getPageNum());
				}
				//System.out.println("removed page: " + page.getPageNum() + "; next = " + page.header.nextPage);
				page.remove();
				fileHeader.setDirty(true);
			}
		}
	}

	/**
	 *  unlinkPages unlinks a set of pages starting at the specified page
	 *  number.
	 *
	 *@param  pageNum       Description of the Parameter
	 *@throws  IOException  if an Exception occurs
	 */
	protected final void unlinkPages(long pageNum) throws IOException {
		unlinkPages(getPage(pageNum));
	}

	/**
	 *  writeValue writes the multi-Paged Value starting at the specified Page.
	 *
	 *@param  page          The starting Page
	 *@param  value         The Value to write
	 *@throws  IOException  if an Exception occurs
	 */
	public final void writeValue(Page page, Value value) throws IOException {
		byte[] data = value.getData();
		PageHeader hdr = page.getPageHeader();
		hdr.setRecordLen(value.getLength());
		hdr.dataLen = fileHeader.workSize;
		if (data.length < hdr.dataLen)
			hdr.dataLen = data.length;
		page.write(data);
	}

	/**
	 *  writeValue writes the multi-Paged Value starting at the specified page
	 *  number.
	 *
	 *@param  page          The starting page number
	 *@param  value         The Value to write
	 *@throws  IOException  if an Exception occurs
	 */
	protected final void writeValue(long page, Value value) throws IOException {
		writeValue(getPage(page), value);
	}

	/**
	 *  FileHeader
	 *
	 *@author     Wolfgang Meier <meier@ifs.tu-darmstadt.de>
	 *@created    26. Mai 2002
	 */

	public abstract class FileHeader {
		private boolean dirty = false;
		long firstFreePage = -1;

		short headerSize;
		private long lastFreePage = -1;
		private short maxKeySize = 256;
		private long pageCount;
		byte pageHeaderSize = 64;
		int pageSize;
		private long recordCount;
		long totalCount;
		int workSize;

		/**  Constructor for the FileHeader object */
		public FileHeader() {
			this(1024, PAGE_SIZE);
		}

		/**
		 *  Constructor for the FileHeader object
		 *
		 *@param  pageCount  Description of the Parameter
		 */
		public FileHeader(long pageCount) {
			this(pageCount, Paged.PAGE_SIZE);
		}

		/**
		 *  Constructor for the FileHeader object
		 *
		 *@param  pageCount  Description of the Parameter
		 *@param  pageSize   Description of the Parameter
		 */
//		public FileHeader(long pageCount, int pageSize) {
//			this(pageCount, pageSize, (byte) 4);
//		}

		/**
		 *  Constructor for the FileHeader object
		 *
		 *@param  pageCount  Description of the Parameter
		 *@param  pageSize   Description of the Parameter
		 *@param  blockSize  Description of the Parameter
		 */
		public FileHeader(long pageCount, int pageSize/*, byte blockSize*/) {
			this.pageSize = pageSize;
			this.pageCount = pageCount;
			totalCount = pageCount;
			headerSize = (short) pageSize;
			calculateWorkSize();
		}

		/**
		 *  Constructor for the FileHeader object
		 *
		 *@param  read             Description of the Parameter
		 *@exception  IOException  Description of the Exception
		 */
		public FileHeader(boolean read) throws IOException {
			if (read)
				read();
		}

		private void calculateWorkSize() {
			workSize = pageSize - pageHeaderSize;
		}

		/**  Decrement the number of records being managed by the file */
		public final synchronized void decRecordCount() {
			recordCount--;
			dirty = true;
		}

		/**
		 *  The first free page in unused secondary space
		 *
		 *@return    The firstFreePage value
		 */
		public final long getFirstFreePage() {
			return firstFreePage;
		}

		/**
		 *  The size of the FileHeader. Usually 1 OS Page
		 *
		 *@return    The headerSize value
		 */
		public final short getHeaderSize() {
			return headerSize;
		}

		/**
		 *  The last free page in unused secondary space
		 *
		 *@return    The lastFreePage value
		 */
		public final long getLastFreePage() {
			return lastFreePage;
		}

		/**
		 *  The maximum number of bytes a key can be. 256 is good
		 *
		 *@return    The maxKeySize value
		 */
		public final short getMaxKeySize() {
			return maxKeySize;
		}

		/**
		 *  The number of pages in primary storage
		 *
		 *@return    The pageCount value
		 */
		public final long getPageCount() {
			return pageCount;
		}

		/**
		 *  The size of a page header. 64 is sufficient
		 *
		 *@return    The pageHeaderSize value
		 */
		public final byte getPageHeaderSize() {
			return pageHeaderSize;
		}

		/**
		 *  The size of a page. Usually a multiple of a FS block
		 *
		 *@return    The pageSize value
		 */
		public final int getPageSize() {
			return pageSize;
		}

		/**
		 *  The number of records being managed by the file (not pages)
		 *
		 *@return    The recordCount value
		 */
		public final long getRecordCount() {
			return recordCount;
		}

		/**
		 *  The number of total pages in the file
		 *
		 *@return    The totalCount value
		 */
		public final long getTotalCount() {
			return totalCount;
		}

		/**
		 *  Gets the workSize attribute of the FileHeader object
		 *
		 *@return    The workSize value
		 */
		public final int getWorkSize() {
			return workSize;
		}

		/**  Increment the number of records being managed by the file */
		public final synchronized void incRecordCount() {
			recordCount++;
			dirty = true;
		}

		/**
		 *  Gets the dirty attribute of the FileHeader object
		 *
		 *@return    The dirty value
		 */
		public final boolean isDirty() {
			return dirty;
		}

		/**
		 *  Description of the Method
		 *
		 *@exception  IOException  Description of the Exception
		 */
		public final synchronized void read() throws IOException {
			raf.seek(0);
			read(raf);
			calculateWorkSize();
			dirty = false;
		}

		/**
		 *  Description of the Method
		 *
		 *@param  aRaf              Description of the Parameter
		 *@exception  IOException  Description of the Exception
		 */
		public void read(RandomAccessFile aRaf) throws IOException {
			headerSize = aRaf.readShort();
			pageSize = aRaf.readInt();
			pageCount = aRaf.readLong();
			totalCount = aRaf.readLong();
			firstFreePage = aRaf.readLong();
			lastFreePage = aRaf.readLong();
			pageHeaderSize = aRaf.readByte();
			maxKeySize = aRaf.readShort();
			recordCount = aRaf.readLong();
		}

		/**
		 *  Sets the dirty attribute of the FileHeader object
		 *
		 *@param  dirty  The new dirty value
		 */
		public final void setDirty(boolean dirty) {
			this.dirty = dirty;
		}

		/**
		 *  The first free page in unused secondary space
		 *
		 *@param  firstFreePage  The new firstFreePage value
		 */
		public final void setFirstFreePage(long firstFreePage) {
			this.firstFreePage = firstFreePage;
			dirty = true;
		}

		/**
		 *  The size of the FileHeader. Usually 1 OS Page
		 *
		 *@param  headerSize  The new headerSize value
		 */
		public final void setHeaderSize(short headerSize) {
			this.headerSize = headerSize;
			dirty = true;
		}

		/**
		 *  The last free page in unused secondary space
		 *
		 *@param  lastFreePage  The new lastFreePage value
		 */
		public final void setLastFreePage(long lastFreePage) {
			this.lastFreePage = lastFreePage;
			dirty = true;
		}

		/**
		 *  The maximum number of bytes a key can be. 256 is good
		 *
		 *@param  maxKeySize  The new maxKeySize value
		 */
		public final void setMaxKeySize(short maxKeySize) {
			this.maxKeySize = maxKeySize;
			dirty = true;
		}

		/**
		 *  The number of pages in primary storage
		 *
		 *@param  pageCount  The new pageCount value
		 */
		public final void setPageCount(long pageCount) {
			this.pageCount = pageCount;
			dirty = true;
		}

		/**
		 *  The size of a page header. 64 is sufficient
		 *
		 *@param  pageHeaderSize  The new pageHeaderSize value
		 */
		public final void setPageHeaderSize(byte pageHeaderSize) {
			this.pageHeaderSize = pageHeaderSize;
			calculateWorkSize();
			dirty = true;
		}

		/**
		 *  The size of a page. Usually a multiple of a FS block
		 *
		 *@param  pageSize  The new pageSize value
		 */
		public final void setPageSize(int pageSize) {
			this.pageSize = pageSize;
			calculateWorkSize();
			dirty = true;
		}

		/**
		 *  The number of records being managed by the file (not pages)
		 *
		 *@param  recordCount  The new recordCount value
		 */
		public final void setRecordCount(long recordCount) {
			this.recordCount = recordCount;
			dirty = true;
		}

		/**
		 *  The number of total pages in the file
		 *
		 *@param  totalCount  The new totalCount value
		 */
		public final void setTotalCount(long totalCount) {
			this.totalCount = totalCount;
			dirty = true;
		}

		/**
		 *  Description of the Method
		 *
		 *@exception  IOException  Description of the Exception
		 */
		public final synchronized void write() throws IOException {
			raf.seek(0);
			write(raf);
			dirty = false;
		}

		/**
		 *  Description of the Method
		 *
		 *@param  aRaf              Description of the Parameter
		 *@exception  IOException  Description of the Exception
		 */
		public void write(RandomAccessFile aRaf) throws IOException {
			aRaf.writeShort(headerSize);
			aRaf.writeInt(pageSize);
			aRaf.writeLong(pageCount);
			aRaf.writeLong(totalCount);
			aRaf.writeLong(firstFreePage);
			aRaf.writeLong(lastFreePage);
			aRaf.writeByte(pageHeaderSize);
			aRaf.writeShort(maxKeySize);
			aRaf.writeLong(recordCount);
		}
	}

	/**
	 *  Page
	 */

	public final class Page implements Comparable {

		/**  The Header for this Page */
		PageHeader header;

		/**  The offset into the file that this page starts */
		private long offset;
		/**  This page number */
		long pageNum;

		private int refCount = 0;
		
		/**  Constructor for the Page object */
		public Page() {
			header = createPageHeader();
		}

		/**
		 *  Constructor for the Page object
		 *
		 *@param  pageNum          Description of the Parameter
		 */
		public Page(long pageNum) {
			this();
			if(pageNum < 0)
				Thread.dumpStack();
			setPageNum(pageNum);
		}

		/**  Description of the Method */
		public void decRefCount() {
			refCount--;
		}

		/**
		 *  Gets the offset attribute of the Page object
		 *
		 *@return    The offset value
		 */
		public long getOffset() {
			return offset;
		}

		/**
		 *  Gets the pageHeader attribute of the Page object
		 *
		 *@return    The pageHeader value
		 */
		public PageHeader getPageHeader() {
			return header;
		}

		/**
		 *  Gets the pageInfo attribute of the Page object
		 *
		 *@return    The pageInfo value
		 */
		public String getPageInfo() {
			return "page: "
				+ pageNum
				+ "; file = "
				+ getFile().getName()
				+ "; address = "
				+ Long.toHexString(offset)
				+ "; page header = "
				+ fileHeader.getPageHeaderSize()
				+ "; data start = "
				+ Long.toHexString(offset + fileHeader.getPageHeaderSize());
		}

		/**
		 *  Gets the pageNum attribute of the Page object
		 *
		 *@return    The pageNum value
		 */
		public long getPageNum() {
			return pageNum;
		}

		/**
		 *  Gets the refCount attribute of the Page object
		 *
		 *@return    The refCount value
		 */
		public int getRefCount() {
			return refCount;
		}

		public int getDataPos() {
			return fileHeader.pageHeaderSize;
		}

		/**  Description of the Method */
		public void incRefCount() {
			refCount++;
		}

		/**
		 *  Description of the Method
		 *
		 *@exception  IOException  Description of the Exception
		 */
		public byte[] read() throws IOException {
			try {
				//System.out.println("reading page="+getPageNum()+" size of header data="+tempHeaderData.length);
				if (raf.getFilePointer() != offset) {
					raf.seek(offset);
				}
				Arrays.fill(tempHeaderData, (byte)0);
				raf.read(tempHeaderData);
	
				// Read in the header
				header.read(tempHeaderData, 0);
				// Read the working data
				final byte[] workData = new byte[header.dataLen];
				//System.out.println("page ="+getPageInfo()+" work data length="+workData.length);
				raf.read(workData);
//				Thread.sleep(5);
				
				return workData;
			} catch(Exception e) {
				LOG.log(Level.FINE ,"error while reading page: " + getPageInfo(), e);
				throw new IOException(e.getMessage());
			}
		}

		/**
		 *  Sets the pageNum attribute of the Page object
		 *
		 *@param  pageNum  The new pageNum value
		 */
		public void setPageNum(long pageNum) {
			this.pageNum = pageNum;
			offset = fileHeader.headerSize + (pageNum * fileHeader.pageSize);
		}

		public void remove() throws IOException {
			write(null);
		}

		final void write(byte[] data) throws IOException {
			//System.out.println(getFile().getName() + " writing page " + pageNum);
			if(data == null)
				// Removed page: fill with 0
				Arrays.fill(tempPageData, (byte)0);
			
			// Write out the header
			header.write(tempPageData, 0);
			header.dirty = false;
			
			if(data != null) {
				if(data.length > fileHeader.workSize) {
					throw new IOException("page: " + getPageInfo() + ": data length too large: " + data.length);				    
				}
				System.arraycopy(data, 0, tempPageData, fileHeader.pageHeaderSize, data.length);
			}
			if (raf.getFilePointer() != offset)
				raf.seek(offset);
			//raf.write(data);
			raf.write(tempPageData);
		}
		
		/* (non-Javadoc)
		 * @see java.lang.Object#equals(java.lang.Object)
		 */
		public boolean equals(Object obj) {
			return ((Page)obj).pageNum == pageNum;
		}
		
		/* (non-Javadoc)
		 * @see java.lang.Comparable#compareTo(java.lang.Object)
		 */
		public int compareTo(Object o) {
			final Page other = (Page)o;
			if(pageNum == other.pageNum)
				return 0;
			else if(pageNum > other.pageNum)
				return 1;
			else
				return -1;
		}

	}

	public abstract class PageHeader {
		int dataLen = 0;
		boolean dirty = false;
		long nextPage = -1;
		private int recordLen = 0;

		private byte status = UNUSED;

		/**  Constructor for the PageHeader object */
		public PageHeader() {
		}

		/**
		 *  Constructor for the PageHeader object
		 *
		 * @param  dis              Description of the Parameter
		 */
		public PageHeader(byte[] data, int offset) throws IOException {
			read(data, offset);
		}

		/**
		 *  The length of the Data
		 *
		 *@return    The dataLen value
		 */
		public final int getDataLen() {
			return dataLen;
		}

		/**
		 *  The next page for this Record (if overflowed)
		 *
		 *@return    The nextPage value
		 */
		public final long getNextPage() {
			return nextPage;
		}

		/**
		 *  The length of the Record's value
		 *
		 *@return    The recordLen value
		 */
		public final int getRecordLen() {
			return recordLen;
		}

		/**
		 *  The status of this page (UNUSED, RECORD, DELETED, etc...)
		 *
		 *@return    The status value
		 */
		public final byte getStatus() {
			return status;
		}

		/**
		 *  Gets the dirty attribute of the PageHeader object
		 *
		 *@return    The dirty value
		 */
		public final boolean isDirty() {
			return dirty;
		}

		/**
		 *  Description of the Method
		 *
		 *@param  dis              Description of the Parameter
		 * @throws IOException
		 */
		public int read(byte[] data, int offset) throws IOException {
			status = data[offset++];
			dataLen = ByteConversion.byteToInt(data, offset);
			offset += 4;
			recordLen = ByteConversion.byteToInt(data, offset);
			offset += 4;
			nextPage = ByteConversion.byteToLong(data, offset);
			offset += 8;
			return offset;
		}

		/**
		 *  The length of the Data
		 *
		 *@param  dataLen  The new dataLen value
		 */
		public final void setDataLen(int dataLen) {
			this.dataLen = dataLen;
			dirty = true;
		}

		/**
		 *  Sets the dirty attribute of the PageHeader object
		 *
		 *@param  dirty  The new dirty value
		 */
		public final void setDirty(boolean dirty) {
			this.dirty = dirty;
		}

		/**
		 *  The next page for this Record (if overflowed)
		 *
		 *@param  nextPage  The new nextPage value
		 */
		public final void setNextPage(long nextPage) {
			this.nextPage = nextPage;
			dirty = true;
		}

		/**
		 *  The length of the Record's value
		 *
		 *@param  recordLen  The new recordLen value
		 */
		public void setRecordLen(int recordLen) {
			this.recordLen = recordLen;
			dirty = true;
		}

		/**
		 *  The status of this page (UNUSED, RECORD, DELETED, etc...)
		 *
		 *@param  status  The new status value
		 */
		public final void setStatus(byte status) {
			this.status = status;
			dirty = true;
		}

		/**
		 *  Description of the Method
		 *
		 *@param  dos              Description of the Parameter
		 */
		public int write(byte[] data, int offset) throws IOException {
			data[offset++] = status;
			ByteConversion.intToByte(dataLen, data, offset);
			offset += 4;
			ByteConversion.intToByte(recordLen, data, offset);
			offset += 4;
			ByteConversion.longToByte(nextPage, data, offset);
			offset += 8;
			dirty = false;
			return offset;
		}
	}
	
}
