package org.exist;

import java.io.File;
import java.io.IOException;
import java.io.UnsupportedEncodingException;

import net.sf.xanswer.storage.map.PersistentMapImplBase;
import net.sf.xanswer.storage.map.PersistentException;
import net.sf.xanswer.storage.map.PersistentFactory;

import org.dbxml.core.DBException;
import org.dbxml.core.data.Value;
import org.dbxml.core.filer.BTreeException;
import org.dbxml.core.filer.Paged;
import org.dbxml.core.filer.ValuesIterator;
import org.exist.storage.store.BFile;

/**
 * This class provides the eXist implementaion of PersistentMap.
 */
public class PersistentMapImplExist extends PersistentMapImplBase /*implements PersistentMap */{
	private BFile myBFile;
	
	PersistentMapImplExist(PersistentFactory factory) {
		super(factory);
	}
	
    protected String getInstanceName() {
    	return "eXist";
    }

	// some notes on caching (and memory using) in Exist:
    // Exist is the chosen one. (And maybe the wrong chosen one). It is the fastest and 
    // the most modest to disk resources while the number of records and the altitude of
    // the tree is ordinary. It loses in speed to Berkeley on large numbers of records
    // (Berkeley can be twice, or even 3 times faster), but disk consumption stays
    // the most modest in any cases (but it hardly may be a big advantage).
    // It is also the most configurable one of this branch of implementations based on
    // the same dbxml core. Another it's advantage is in the most weak relations between
    // classes and packages, which may be explained by author's JDepend favouring. :) 
    // Speaking about memory consumption and cache, its caching mechanizms seemes to be
    // the most reasonable of all 3 brothers (which rely on WeakReference mechanisms).
    // Here the LRDCache(Least Reference Density) provided by author is used. This
    // cache manages its Cacheable items saving by calling their sync() method. That is
    // a very convinient decision, but, unfortunately, it may cause a horrible bug because
    // of returning ordinary, desne-references to its items by the Cache.get() method.
    // And, unfortunately, this bug was really discovered by my tests (certanly it was
    // immediately fixed). But we are still not protected from it's appearence another
    // time at another place, if code modifications would be done by not-devoted person
    // (as I am) :) or by mistake. Weak references in other implementations at least
    // give us reminding and feeling more secure, as we know they could be nulled at
    // any time (by gc).
    // Concrete speaking about classes implementation, it is necessary to mention their
    // BFiler implementation. Opposite to Paged in this implementation, it uses cache
    // and free space managing mechanisms, which provide, I suppose, such good disk
    // space consumption. The author did not limit the Filer duties with only mapping 
    // pointers comes from leafs of BTree to pages in file, he provided some more
    // efficient mechanism of managing space and memory used by values.
    // And also he left this mechanism to be adjustable, unlike other branch impls do.
    // Paged implementation is left very simple, no caching is made on the level
    // of managing pages, it relies on subclasses to provide caching. (It is not wholy
    // true, because current page is always cached in Paged, surely).
    // Speaking of synchronization, Exist in this version is not made for concurrent
    // access. Paged provides syncronization of pages by locking their page header. 
    // And BFile provides a lock for their OverflowPages. It is the only minimum can
    // be done on synchronization, but it provides a good speed increase. 
    // Of course, synchronization is to be done anyway, and we were informed, that in 
    // current version of Exist the author has fixed this.
    // Speaking about the BTree algorithm, we have noticed some changes provided to the
    // standard implementation in Xindice and DbXML. For example, separators are not
    // used and replaced by promoting the actual value found in the leaf. Nodes do not
    // write any changes to the disk each time and rely only on the cache to do it
    // in due time. Of course, this could affect (and does affect) the loss of data
    // in extra cases. But the cache implementation is improving every day :) and by
    // now it is providing a timer mechanism (also is configurable), which flushes any
    // dirty data at certain time periods.
    // Well, I guess "it's all I can say about the war in Vietnam". :))
    // But further versions are still under investigation. :)
    
    // TODO: provide a balancing mechanism for the b-tree algorithm and test it.
    
    protected void createDatabase(File location, int pageSize, int buffersCount) throws PersistentException {
		Paged.setPageSize(pageSize); // in bytes
		// in Cacheable items (Pages)(Nodes) - 256 by default in both caches:
	    myBFile = new BFile(location, buffersCount/*BTree Cache*/, 256/*BFile DataCache - 256 by default*/);
	    if (!myBFile.exists()) {
		    try {
               myBFile.create();
           } catch (DBException e) {
               throw new PersistentException(e);
           } 
	    }
    }

    public void openDatabase() throws PersistentException {
        try {
        	myBFile.open();
        } catch (DBException e) {
            throw new PersistentException(e);
        }
    }

    public void closeDatabase() throws PersistentException {
        try {
        	myBFile.flush();
        	myBFile.close();
//        	myKeysIterator = null;
        } catch (DBException e) {
            throw new PersistentException(e);
        }
    }

	protected void putBytes(String key, byte[] value) throws PersistentException {
//    	    this.removeBytes(key); - remove is not necessary, we have an owerwrite flag here
			myBFile.put(new Value(key), value, true);
//			System.err.println("put key: " + key + " value: " + value);
	}

	protected byte[] getBytes(String key) throws PersistentException {
		byte[] result = new byte[0];
		Value value = myBFile.get(new Value(key));
		if (value != null) {
			result = value.getData(); 		    
		}
//		if ((result != null) && (result.length != 0)) System.err.println("get key: " + key + " value: " + result);
	    return result;
	}

	protected void removeBytes(String key) throws PersistentException {
		myBFile.remove(new Value(key));
	}

    protected StringIterator getIterator() throws PersistentException {
        return new StringIteratorImplExist(myBFile.iterator());
    }
	
    protected long getNumberOfRecords() throws PersistentException {
        long result = 0;
        try {
            result = myBFile.getKeys().size();
        } catch (BTreeException e) {
            throw new PersistentException(e);
        } catch (IOException e) {
            throw new PersistentException(e);
        }
        return result;
    }
    
    private class StringIteratorImplExist implements StringIterator {
        ValuesIterator myIterator;
        
        protected StringIteratorImplExist(ValuesIterator iterator) {
            myIterator = iterator;
        }

        public boolean hasNext() throws PersistentException {
            boolean result = false;
            try {
                result = myIterator.hasNext();
            } catch (BTreeException e) {
    			throw new PersistentException(e);
            }
            return result;
        }

        public String next() throws PersistentException {
            String result = "";
            try {
                Value nextValue = myIterator.next();
                if (nextValue != null) {
    				result = new String(nextValue.getData(), "UTF-8");                    
                }
            } catch (BTreeException e) {
    			throw new PersistentException(e);
            } catch (UnsupportedEncodingException e) {
    			throw new PersistentException(e);
            }            
            return result;
        }

    }

}
